package com.sun.jmx.mbeanserver;

import javax.management.InstanceNotFoundException;
import java.util.Collections;
import java.util.HashSet;
import javax.management.QueryExp;
import javax.management.InstanceAlreadyExistsException;
import java.util.ArrayList;
import java.util.logging.Level;
import com.sun.jmx.defaults.JmxProperties;
import javax.management.RuntimeOperationsException;
import java.util.HashMap;
import javax.management.ObjectName;
import javax.management.DynamicMBean;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.Map;

public class Repository
{
    private final Map<String, Map<String, NamedObject>> domainTb;
    private volatile int nbElements;
    private final String domain;
    private final ReentrantReadWriteLock lock;
    
    private void addAllMatching(Map<String, NamedObject> map, Set<NamedObject> set, Repository.ObjectNamePattern objectNamePattern) {
        synchronized (map) {
            for (NamedObject namedObject : map.values()) {
                if (objectNamePattern.matchKeys(namedObject.getName())) {
                    set.add(namedObject);
                }
                else {
                    continue;
                }
            }
        }
    }
    
    private void addNewDomMoi(DynamicMBean dynamicMBean, String s, ObjectName objectName, Repository.RegistrationContext registrationContext) {
        HashMap<String, NamedObject> hashMap = new HashMap();
        this.addMoiToTb(dynamicMBean, objectName, objectName.getCanonicalKeyPropertyListString(), hashMap, registrationContext);
        this.domainTb.put(s, hashMap);
        ++this.nbElements;
    }
    
    private void registering(Repository.RegistrationContext registrationContext) {
        if (registrationContext == null) {
            return;
        }
        try {
            registrationContext.registering();
        }
        catch (RuntimeOperationsException ex) {
            throw ex;
        }
        catch (RuntimeException e) {
            throw new RuntimeOperationsException(e);
        }
        return;
    }
    
    private void unregistering(Repository.RegistrationContext registrationContext, ObjectName obj) {
        if (registrationContext == null) {
            return;
        }
        try {
            registrationContext.unregistered();
        }
        catch (Exception thrown) {
            JmxProperties.MBEANSERVER_LOGGER.log(Level.FINE, "Unexpected exception while unregistering " + obj, thrown);
        }
        return;
    }
    
    private void addMoiToTb(DynamicMBean dynamicMBean, ObjectName objectName, String s, Map<String, NamedObject> map, Repository.RegistrationContext registrationContext) {
        this.registering(registrationContext);
        map.put(s, new NamedObject(objectName, dynamicMBean));
    }
    
    private NamedObject retrieveNamedObject(ObjectName objectName) {
        if (objectName.isPattern()) {
            return null;
        }
        String s = objectName.getDomain().intern();
        if (s.length() == 0) {
            s = this.domain;
        }
        Map<Object, Object> map = (Map)this.domainTb.get(s);
        if (map == null) {
            return null;
        }
        return (NamedObject)map.get(objectName.getCanonicalKeyPropertyListString());
    }
    
    public Repository(String s) {
        this(s, true);
    }
    
    public Repository(String s, boolean fair) {
        this.nbElements = 0;
        this.lock = new ReentrantReadWriteLock(fair);
        this.domainTb = new HashMap(5);
        if (s != null && s.length() != 0) {
            this.domain = s.intern();
        }
        else {
            this.domain = "DefaultDomain";
        }
        this.domainTb.put(this.domain, new HashMap());
    }
    
    public String[] getDomains() {
        this.lock.readLock().lock();
        ArrayList list;
        try {
            list = new ArrayList(this.domainTb.size());
            for (Map.Entry entry : this.domainTb.entrySet()) {
                Map<Object, Object> map = (Map)entry.getValue();
                if (map != null && map.size() != 0) {
                    list.add(entry.getKey());
                }
                else {
                    continue;
                }
            }
        }
        finally {
            this.lock.readLock().unlock();
        }
        return (String[])list.toArray(new String[list.size()]);
    }
    
    public void addMBean(DynamicMBean dynamicMBean, ObjectName objectName, Repository.RegistrationContext registrationContext) throws InstanceAlreadyExistsException {
        if (JmxProperties.MBEANSERVER_LOGGER.isLoggable(Level.FINER)) {
            JmxProperties.MBEANSERVER_LOGGER.logp(Level.FINER, Repository.class.getName(), "addMBean", "name = " + objectName);
        }
        String s = objectName.getDomain().intern();
        if (s.length() == 0) {
            objectName = Util.newObjectName(this.domain + objectName.toString());
        }
        int n;
        if (s == this.domain) {
            n = true;
            s = this.domain;
        }
        else {
            n = false;
        }
        if (objectName.isPattern()) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Repository: cannot add mbean for pattern name " + objectName.toString()));
        }
        this.lock.writeLock().lock();
        try {
            if (n == 0 && s.equals("JMImplementation") && this.domainTb.containsKey("JMImplementation")) {
                throw new RuntimeOperationsException(new IllegalArgumentException("Repository: domain name cannot be JMImplementation"));
            }
            Map<String, NamedObject> map = (Map)this.domainTb.get(s);
            if (map == null) {
                this.addNewDomMoi(dynamicMBean, s, objectName, registrationContext);
                return;
            }
            String canonicalKeyPropertyListString = objectName.getCanonicalKeyPropertyListString();
            if ((NamedObject)map.get(canonicalKeyPropertyListString) != null) {
                throw new InstanceAlreadyExistsException(objectName.toString());
            }
            ++this.nbElements;
            this.addMoiToTb(dynamicMBean, objectName, canonicalKeyPropertyListString, map, registrationContext);
        }
        finally {
            this.lock.writeLock().unlock();
        }
        return;
    }
    
    public boolean contains(ObjectName obj) {
        if (JmxProperties.MBEANSERVER_LOGGER.isLoggable(Level.FINER)) {
            JmxProperties.MBEANSERVER_LOGGER.logp(Level.FINER, Repository.class.getName(), "contains", " name = " + obj);
        }
        this.lock.readLock().lock();
        try {
            return this.retrieveNamedObject(obj) != null;
        }
        finally {
            this.lock.readLock().unlock();
        }
    }
    
    public DynamicMBean retrieve(ObjectName obj) {
        if (JmxProperties.MBEANSERVER_LOGGER.isLoggable(Level.FINER)) {
            JmxProperties.MBEANSERVER_LOGGER.logp(Level.FINER, Repository.class.getName(), "retrieve", "name = " + obj);
        }
        this.lock.readLock().lock();
        try {
            NamedObject retrieveNamedObject = this.retrieveNamedObject(obj);
            if (retrieveNamedObject == null) {
                return null;
            }
            return retrieveNamedObject.getObject();
        }
        finally {
            this.lock.readLock().unlock();
        }
    }
    
    public Set<NamedObject> query(ObjectName objectName, QueryExp queryExp) {
        HashSet<NamedObject> set = new HashSet();
        ObjectName wildcard;
        if (objectName == null || objectName.getCanonicalName().length() == 0 || objectName.equals(ObjectName.WILDCARD)) {
            wildcard = ObjectName.WILDCARD;
        }
        else {
            wildcard = objectName;
        }
        this.lock.readLock().lock();
        try {
            if (!wildcard.isPattern()) {
                NamedObject retrieveNamedObject = this.retrieveNamedObject(wildcard);
                if (retrieveNamedObject != null) {
                    set.add(retrieveNamedObject);
                }
                return set;
            }
            else {
                if (wildcard == ObjectName.WILDCARD) {
                    Iterator<Map<String, NamedObject>> iterator2 = this.domainTb.values().iterator();
                    while (iterator2.hasNext()) {
                        set.addAll(((Map)iterator2.next()).values());
                    }
                    return set;
                }
                int n = wildcard.getCanonicalKeyPropertyListString().length() == 0;
                Object o = n ? null : new ObjectNamePattern(wildcard);
                if (wildcard.getDomain().length() == 0) {
                    Map<String, NamedObject> map3 = (Map)this.domainTb.get(this.domain);
                    if (!(n == 0)) {
                        set.addAll(map3.values());
                    }
                    else {
                        this.addAllMatching(map3, set, o);
                    }
                    return set;
                }
                else if (!wildcard.isDomainPattern()) {
                    Map<String, NamedObject> map2 = (Map)this.domainTb.get(wildcard.getDomain());
                    if (map2 == null) {
                        return Collections.emptySet();
                    }
                    if (!(n == 0)) {
                        set.addAll(map2.values());
                    }
                    else {
                        this.addAllMatching(map2, set, o);
                    }
                    return set;
                }
                else {
                    String domain = wildcard.getDomain();
                    for (String s : this.domainTb.keySet()) {
                        if (Util.wildmatch(s, domain)) {
                            Map<String, NamedObject> map = (Map)this.domainTb.get(s);
                            if (!(n == 0)) {
                                set.addAll(map.values());
                            }
                            else {
                                this.addAllMatching(map, set, o);
                            }
                        }
                        else {
                            continue;
                        }
                    }
                    return set;
                }
            }
        }
        finally {
            this.lock.readLock().unlock();
        }
    }
    
    public void remove(ObjectName obj, Repository.RegistrationContext registrationContext) throws InstanceNotFoundException {
        if (JmxProperties.MBEANSERVER_LOGGER.isLoggable(Level.FINER)) {
            JmxProperties.MBEANSERVER_LOGGER.logp(Level.FINER, Repository.class.getName(), "remove", "name = " + obj);
        }
        String s = obj.getDomain().intern();
        if (s.length() == 0) {
            s = this.domain;
        }
        this.lock.writeLock().lock();
        try {
            Map<Object, Object> map = (Map)this.domainTb.get(s);
            if (map == null) {
                throw new InstanceNotFoundException(obj.toString());
            }
            if (map.remove(obj.getCanonicalKeyPropertyListString()) == null) {
                throw new InstanceNotFoundException(obj.toString());
            }
            --this.nbElements;
            if (map.isEmpty()) {
                this.domainTb.remove(s);
                if (s == this.domain) {
                    this.domainTb.put(this.domain, new HashMap());
                }
            }
            this.unregistering(registrationContext, obj);
        }
        finally {
            this.lock.writeLock().unlock();
        }
    }
    
    public Integer getCount() {
        return Integer.valueOf(this.nbElements);
    }
    
    public String getDefaultDomain() {
        return this.domain;
    }
    
    private static final class ObjectNamePattern
    {
        private final String[] keys;
        private final String[] values;
        private final String properties;
        private final boolean isPropertyListPattern;
        private final boolean isPropertyValuePattern;
        public final ObjectName pattern;
        
        public ObjectNamePattern(ObjectName objectName) {
            this(objectName.isPropertyListPattern(), objectName.isPropertyValuePattern(), objectName.getCanonicalKeyPropertyListString(), objectName.getKeyPropertyList(), objectName);
        }
        
        ObjectNamePattern(boolean isPropertyListPattern, boolean isPropertyValuePattern, String properties, Map<String, String> map, ObjectName pattern) {
            this.isPropertyListPattern = isPropertyListPattern;
            this.isPropertyValuePattern = isPropertyValuePattern;
            this.properties = properties;
            int size = map.size();
            this.keys = new String[size];
            this.values = new String[size];
            int n = 0;
            for (Map.Entry entry : map.entrySet()) {
                this.keys[n] = (String)entry.getKey();
                this.values[n] = (String)entry.getValue();
                ++n;
            }
            this.pattern = pattern;
        }
        
        public boolean matchKeys(ObjectName objectName) {
            if (this.isPropertyValuePattern && !this.isPropertyListPattern && objectName.getKeyPropertyList().size() != this.keys.length) {
                return false;
            }
            if (this.isPropertyValuePattern || this.isPropertyListPattern) {
                int i = this.keys.length - 1;
                while (i >= 0) {
                    String keyProperty = objectName.getKeyProperty(this.keys[i]);
                    if (keyProperty == null) {
                        return false;
                    }
                    if (this.isPropertyValuePattern && this.pattern.isPropertyValuePattern(this.keys[i])) {
                        if (!Util.wildmatch(keyProperty, this.values[i])) {
                            return false;
                        }
                    }
                    else if (!keyProperty.equals(this.values[i])) {
                        return false;
                    }
                    --i;
                }
                return true;
            }
            return objectName.getCanonicalKeyPropertyListString().equals(this.properties);
        }
    }
    
    public interface RegistrationContext
    {
        public abstract void registering();
        
        public abstract void unregistered();
    }
}
